import pytest

from b_lambda_layer_common.exceptions.container.bad_request_error import BadRequestError
from b_lambda_layer_common.util.field_validator import FieldValidator


def test_FUNC_max_length_WITH_too_long_string_EXPECT_exception_raised() -> None:
    """
    Check whether the validation succeeded.

    :return: No return.
    """
    with pytest.raises(BadRequestError):
        FieldValidator.max_length('RandomString', 5)


def test_FUNC_max_length_WITH_null_string_EXPECT_exception_not_raised() -> None:
    """
    Check whether the validation succeeded.

    :return: No return.
    """
    FieldValidator.max_length(None, 5)


def test_FUNC_max_length_WITH_good_string_EXPECT_exception_not_raised() -> None:
    """
    Check whether the validation succeeded.

    :return: No return.
    """
    FieldValidator.max_length('RandomString', 50, False)


def test_FUNC_country_iso_3_WITH_not_iso_string_EXPECT_exception_raised() -> None:
    """
    Check whether the validation succeeded.

    :return: No return.
    """
    with pytest.raises(BadRequestError):
        FieldValidator.country_iso_3('RandomString')


def test_FUNC_country_iso_3_WITH_null_string_EXPECT_exception_not_raised() -> None:
    """
    Check whether the validation succeeded.

    :return: No return.
    """
    FieldValidator.country_iso_3(None)


def test_FUNC_country_iso_3_WITH_good_string_EXPECT_exception_not_raised() -> None:
    """
    Check whether the validation succeeded.

    :return: No return.
    """
    FieldValidator.country_iso_3('LTU')


def test_FUNC_string_date_WITH_not_date_like_string_EXPECT_exception_raised() -> None:
    """
    Check whether the validation succeeded.

    :return: No return.
    """
    with pytest.raises(BadRequestError):
        FieldValidator.string_date('NotADate')


def test_FUNC_string_date_WITH_null_string_EXPECT_exception_not_raised() -> None:
    """
    Check whether the validation succeeded.

    :return: No return.
    """
    FieldValidator.string_date(None)


def test_FUNC_string_date_WITH_good_string_EXPECT_exception_not_raised() -> None:
    """
    Check whether the validation succeeded.

    :return: No return.
    """
    FieldValidator.string_date('2020-01-01')


def test_FUNC_string_email_WITH_non_email_string_EXPECT_exception_raised() -> None:
    """
    Check whether the validation succeeded.

    :return: No return.
    """
    with pytest.raises(BadRequestError):
        FieldValidator.string_email('dude.com')


def test_FUNC_string_email_WITH_null_string_EXPECT_exception_not_raised() -> None:
    """
    Check whether the validation succeeded.

    :return: No return.
    """
    FieldValidator.string_email(None)
    
    
def test_FUNC_string_email_WITH_good_string_EXPECT_exception_not_raised() -> None:
    """
    Check whether the validation succeeded.

    :return: No return.
    """
    FieldValidator.string_email('dude@email.com')



from typing import Any

from ordered_set import OrderedSet

from b_lambda_layer_common.util.decimal_encoder import DecimalEncoder


class DynamoDBEncoder(DecimalEncoder):
    def default(self, o: Any):
        if isinstance(o, (set, OrderedSet)):
            return list(o)
        return super(DynamoDBEncoder, self).default(o)



from b_lambda_layer_common.util.skip_invocation import skip_invocation


def test_FUNC_with_default_skip_invocation_EXPECT_successful_return():
    """
    Check whether decorated function returns expected result when no parameters
    to decorator are passed.

    :return: No return.
    """
    item = {'heartbeat': True}
    context = ''

    @skip_invocation()
    def handler(event, context):
        return event

    response = handler(item, context)
    assert response is not None

def test_FUNC_with_bad_determinator_type_EXPECT_successful_return():
    """
    Check whether decorated function returns expected result when wrong
    determinator type is provided as parameter.

    :return: No return.
    """
    item = {'heartbeat': True}
    context = ''

    @skip_invocation(determinator={'test': 'test'})
    def handler(event, context):
        return event

    response = handler(item, context)
    assert response is not None

def test_FUNC_with_string_determinator_EXPECT_none_return():
    """
    Check whether decorated function is not invoked when determinator of type 'str'
    is provided and this key exists in event 'dict'.

    :return: No return.
    """
    item = {'heartbeat': True}
    context = ''

    @skip_invocation(determinator='heartbeat')
    def handler(event, context):
        return event

    response = handler(item, context)
    assert response is None

def test_FUNC_with_string_determinator_EXPECT_successful_return():
    """
    Check whether decorated function is invoked when determinator of type 'str'
    is provided and this key does not exist in event 'dict'.

    :return: No return.
    """
    item = {'heartbeat': True}
    context = ''

    @skip_invocation(determinator='test')
    def handler(event, context):
        return event

    response = handler(item, context)
    assert response is not None

def test_FUNC_with_bool_determinator_EXPECT_none_return():
    """
    Check whether decorated function is not invoked when a determinator of type 'bool'
    is provided and its value is True.

    :return: No return.
    """
    item = {'heartbeat': True}
    context = ''

    @skip_invocation(determinator=True)
    def handler(event, context):
        return event

    response = handler(item, context)
    assert response is None

def test_FUNC_with_bool_determinator_EXPECT_successful_return():
    """
    Check whether decorated function is invoked when a determinator of type 'bool'
    is provided and its value is False.

    :return: No return.
    """
    item = {'heartbeat': True}
    context = ''

    @skip_invocation(determinator=False)
    def handler(event, context):
        return event

    response = handler(item, context)
    assert response is not None

def test_FUNC_with_callable_determinator_EXPECT_none_return():
    """
    Check whether decorated function is not invoked when a 'callable' determinator is provided
    and it returns 'bool' of value True.

    :return: No return.
    """
    item = {'heartbeat': True}
    context = ''

    is_heartbeat = lambda e, c: 'heartbeat' in e

    @skip_invocation(determinator=is_heartbeat)
    def handler(event, context):
        return event

    response = handler(item, context)
    assert response is None

def test_FUNC_with_callable_determinator_EXPECT_successful_return():
    """
    Check whether decorated function is invoked when a 'callable' determinator is provided
    and it returns 'bool' of value False.

    :return: No return.
    """
    item = {'heartbeat': True}
    context = ''

    is_ping = lambda e, c: 'ping' in e

    @skip_invocation(determinator=is_ping)
    def handler(event, context):
        return event

    response = handler(item, context)
    assert response is not None



from b_lambda_layer_common.pynamo_db.kms_attribute import KmsAttribute


def test_FUNC_serialize_WITH_null_value_EXPECT_serialzied() -> None:
    """
    Test whether the serialization works.

    :return: No return.
    """
    att = KmsAttribute(None, 'arn')
    att.serialize(None)


def test_FUNC_deserialize_WITH_null_value_EXPECT_deserialzied() -> None:
    """
    Test whether the serialization works.

    :return: No return.
    """
    att = KmsAttribute(None, 'arn')
    att.deserialize(None)



from typing import Optional

from b_lambda_layer_common.exceptions.http_exception import HttpException


class MalformedPermissionError(HttpException):
    def __init__(self, message: Optional[str] = None):
        super().__init__(message)

    @staticmethod
    def http_code() -> int:
        return 400

    @staticmethod
    def identifier() -> str:
        return 'B_MALFORMED_PERMISSION'

    @staticmethod
    def description() -> str:
        return (
            'The client provided a malformed permission object which does not exist in the server or can not be '
            'understood by the server.'
        )



import json
import math
from itertools import islice
from typing import Any, Iterable, List


class ListBytesBatcher:
    def __init__(self, items: List[Any], batch_size: int) -> None:
        """
        Split the given list of any items into designated size batches.

        :param items: A list of any items that should be split into batches.
        :param batch_size: Size of batch in bytes. Positive non-zero value.

        return No return.
        """

        self.items = items
        self.batch_size = batch_size

    def iterate_batches(self) -> Iterable:
        """
        Split given list into batches not exceeding batch size.

        :return: Iterator of batches.
        """

        if self.batch_size <= 0:
            raise ValueError(f'Unexpected batch size value. Expected positive non-zero integer, got: {self.batch_size}')

        if not isinstance(self.items, list):
            raise TypeError(f'Unexpected data type of items. Expected: {type([])} got: {type(self.items)}.')

        slices = math.ceil(ListBytesBatcher.size(self.items)/self.batch_size)
        batch_size = math.ceil(len(self.items)/slices)
        iterable = iter(self.items)
        while batch := list(islice(iterable, batch_size)):
            yield batch

    @staticmethod
    def size(items: List[Any]) -> int:
        """
        Calculate a number of bytes of given list.

        :param items: A list of any items.

        :return: Number of bytes of given list in UTF-8 encoded form.
        """

        return len(json.dumps(items).encode('utf-8'))



from aws_cdk.aws_lambda import Function, Code, Runtime
from aws_cdk.core import Stack
from b_aws_testing_framework.tools.cdk_testing.testing_stack import TestingStack

from b_lambda_layer_common.layer import Layer


class Function2(Function):
    """
    Function that does not really test anything.
    """
    def __init__(self, scope: Stack):
        super().__init__(
            scope=scope,
            id=f'{TestingStack.global_prefix()}TestingFunction2',
            code=Code.from_inline('def handler(*args, **kwargs): return 200'),
            handler='index.handler',
            runtime=Runtime.PYTHON_3_6,
            layers=[Layer(scope=scope, name=f'{TestingStack.global_prefix()}TestingLayer2')]
        )



import functools
from typing import Any, Callable, Dict, Optional, Union


def skip_invocation(determinator: Optional[Union[str, bool, Callable[[Dict[str, Any], Any], bool]]] = None) -> Union[Callable, None]:
    """
    A decorator which allows to skip decorated function's execution if certain conditions are met.
    By default it returns decorated function's execution if no arguments are provided.
    If a determinator of type 'str' is provided, the decorator checks whether this key exists function's event 'dict' as parses result as 'bool' type.
    If a determinator of type 'bool' is provided, the decorator parses its value.
    If a callable object is provided as determinator, the decorator parses its value as 'bool'.
    Callable should accept 2 arguments - Lambda function's event and context.
    
    If any of the above evaluate to 'bool' of value True, the decorated function is not executed.

    :param determinator: An object which is used to evaluate whether to execute the decorated function.

    :return: Decorated function or None.


    Examples
    ========

    When no parameter is passed as decorator's determinator.

    >>> item = {'heartbeat': True}
    >>> context = ''
    >>> @skip_invocation()
    ... def handler(event, context):
    ...     return event
    >>> handler(item, context)
    {'heartbeat': True}

    When a 'bool' of value True is provided (decorated function does not invoke).

    >>> item = {'heartbeat': True}
    >>> context = ''
    >>> @skip_invocation(determinator=True)
    ... def handler(event, context):
    ...     return event
    >>> handler(item, context)
    >>>

    When a 'str' is provided which exists in event's 'dict' (decorated function does not invoke).

    >>> item = {'heartbeat': True}
    >>> context = ''
    >>> @skip_invocation(determinator="heartbeat")
    ... def handler(event, context):
    ...     return event
    >>> handler(item, context)
    >>>

    When a callable is provided which evaluates to True.
    Callable should take 2 arguments - Lambda function's event and context.

    >>> item = {'heartbeat': True}
    >>> context = ''
    >>> is_heartbeat = lambda e, c: "heartbeat" in e
    >>> @skip_invocation(determinator=is_heartbeat)
    ... def handler(event, context):
    ...     return event
    >>> handler(item, context)
    >>>

    """
    def wrapper(func):
        @functools.wraps(func)
        def wrapped_f(*args, **kwargs):
            skip = False
            event: Dict[str, Any] = args[0]
            context: Any = args[1]
            if isinstance(determinator, str):
                skip = bool(event.get(determinator, False))
            elif isinstance(determinator, bool):
                skip = determinator
            elif callable(determinator):
                skip = bool(determinator(event, context))
            
            if skip:
                return
            return func(*args, **kwargs)
        return wrapped_f
    return wrapper


from aws_cdk.aws_lambda import Function, Code, Runtime
from aws_cdk.core import Stack
from b_aws_testing_framework.tools.cdk_testing.testing_stack import TestingStack
from b_cfn_lambda_layer.package_version import PackageVersion

from b_lambda_layer_common.layer import Layer


class Function1(Function):
    """
    Function that allows us to test whether correct versions are installed.
    """
    def __init__(self, scope: Stack):
        super().__init__(
            scope=scope,
            id=f'{TestingStack.global_prefix()}TestingFunction1',
            code=Code.from_inline(
                'from b_lambda_layer_common import api_gateway\n'
                'from b_lambda_layer_common import exceptions\n'
                'from b_lambda_layer_common import ssm\n'
                'from b_lambda_layer_common import util\n'
                'import boto3\n'
                'import botocore\n'
                '\n\n'
                'def handler(*args, **kwargs):\n'
                '    return dict(\n'
                '        Boto3Version=boto3.__version__,\n'
                '        BotocoreVersion=botocore.__version__,\n'
                '    )'
                '\n'
            ),
            handler='index.handler',
            runtime=Runtime.PYTHON_3_6,
            layers=[
                Layer(
                    scope=scope,
                    name=f'{TestingStack.global_prefix()}TestingLayer1',
                    dependencies={
                        'boto3': PackageVersion.from_string_version('1.16.35'),
                        'botocore': PackageVersion.from_string_version('1.19.35'),
                    }
                )
            ]
        )



from aws_cdk.aws_lambda import Function, Code, Runtime
from aws_cdk.core import Stack
from b_aws_testing_framework.tools.cdk_testing.testing_stack import TestingStack

from b_lambda_layer_common.layer import Layer


class Function4(Function):
    """
    Function which allows us to test whether 'skip_invocation' decorator works as expected.
    """
    def __init__(self, scope: Stack):
        super().__init__(
            scope=scope,
            id=f'{TestingStack.global_prefix()}TestingFunction4',
            code=Code.from_inline(
                'from b_lambda_layer_common.util.skip_invocation import skip_invocation\n'
                '\n\n'
                '@skip_invocation(determinator="heartbeat")\n'
                'def handler(event, context):\n'
                '    return event'
            ),
            handler='index.handler',
            runtime=Runtime.PYTHON_3_6,
            layers=[Layer(scope=scope, name=f'{TestingStack.global_prefix()}TestingLayer4')]
        )



from b_lambda_layer_common.cache.in_memory_data_cache import InMemoryDataCache


def test_FUNC_use_cache_WITH_cache_empty_EXPECT_cache_saved() -> None:
    """
    Test the default value for message.

    :return: No return.
    """
    cache = InMemoryDataCache()

    data = cache.use_cache(
        pointer='123',
        func=lambda: '123'
    )

    assert data == '123'



from typing import Any

from pynamodb.attributes import MapAttribute

from b_lambda_layer_common.util.dynamodb_encoder import DynamoDBEncoder


# TODO move to pynamo_db directory.
class PynamoDBEncoder(DynamoDBEncoder):
    def default(self, o: Any):
        if isinstance(o, MapAttribute):
            return o.as_dict()
        return super(PynamoDBEncoder, self).default(o)



from os.path import dirname, abspath
from typing import Optional, List, Dict

from aws_cdk.aws_lambda import Runtime
from aws_cdk.core import Stack, DockerImage
from b_cfn_lambda_layer.lambda_layer import LambdaLayer
from b_cfn_lambda_layer.package_version import PackageVersion


class Layer(LambdaLayer):
    def __init__(
            self,
            scope: Stack,
            name: str,
            additional_pip_install_args: Optional[str] = None,
            dependencies: Optional[Dict[str, PackageVersion]] = None,
            docker_image: Optional[DockerImage] = None,
    ) -> None:
        super().__init__(
            scope,
            name,
            source_path=self.get_source_path(),
            code_runtimes=self.runtimes(),
            include_source_path_directory=True,
            additional_pip_install_args=additional_pip_install_args,
            dependencies=dependencies,
            docker_image=docker_image
        )

    @staticmethod
    def get_source_path() -> str:
        return dirname(abspath(__file__))

    @staticmethod
    def runtimes() -> Optional[List[Runtime]]:
        return [
            Runtime.PYTHON_3_6,
            Runtime.PYTHON_3_7,
            Runtime.PYTHON_3_8
        ]



import pytest


def handler(*args, **kwargs):
    """
    Programmatic access to run tests with pytest.

    WARNING. Do not delete this file. It is also used for lambda functions to run unit tests.
    """
    return dict(ExitCode=pytest.main([]))



from typing import Optional

from b_lambda_layer_common.exceptions.http_exception import HttpException


class BadRequestError(HttpException):
    def __init__(self, message: Optional[str] = None):
        super().__init__(message)

    @staticmethod
    def http_code() -> int:
        return 400

    @staticmethod
    def identifier() -> str:
        return 'B_BAD_REQUEST'

    @staticmethod
    def description() -> str:
        return (
            'The server cannot or will not process the request due to something that is perceived to be a client '
            'error (e.g., malformed request syntax, invalid request message framing, or deceptive request routing).'
        )



from aws_cdk.aws_lambda import Function, Code, Runtime
from aws_cdk.core import Stack, Duration
from b_aws_testing_framework.tools.cdk_testing.testing_stack import TestingStack
from b_cfn_lambda_layer.package_version import PackageVersion

from b_lambda_layer_common.layer import Layer
from b_lambda_layer_common_test.unit import root


class FunctionWithUnitTests(Function):
    """
    Function that lets us run unit tests inside lambda function. We want to run unit
    tests both locally and remotely.
    """
    def __init__(self, scope: Stack):
        super().__init__(
            scope=scope,
            id=f'{TestingStack.global_prefix()}FunctionWithUnitTests',
            code=Code.from_asset(root),
            handler='handler.handler',
            runtime=Runtime.PYTHON_3_8,
            timeout=Duration.minutes(5),
            memory_size=512,
            layers=[
                Layer(
                    scope=scope,
                    name=f'{TestingStack.global_prefix()}TestingLayerWithUnitTests',
                    dependencies={
                        # These dependencies are required for running unit tests inside lambda functions.
                        # Pytest is used for running actual unit tests.
                        'pytest': PackageVersion.from_string_version('6.2.5'),
                        # Pook is used for HTTP mocking, therefore it is also needed here.
                        'pook': PackageVersion.from_string_version('1.0.1'),
                        # Not sure about this dependency. Lambda runtime throws errors if its missing.
                        'aws-cdk.core': PackageVersion.from_string_version('1.99.0'),
                        # This dependency should be installed with 'pook' since it depends on 'jsonschema' which depends on this.
                        # For some reason it doesn't.
                        # Tests would fail with import error otherwise.
                        'importlib-resources': PackageVersion.from_string_version('5.4.0')
                    }
                )
            ]
        )



from b_lambda_layer_common.util.list_bytes_batcher import ListBytesBatcher


def test_FUNC_iterate_batches_WITH_dummy_list_EXPECT_successful_split() -> None:
    """
    Checks do ListBatcher correctly splits a given list into desired size batches. 

    :return: No return.
    """

    items = [item for item in range(1000*1000)]
    batch_size = 256*1024
    # Split list of items into batches not exceeding batch size.
    pool = ListBytesBatcher(items=items, batch_size=batch_size)

    # Check size of each batch in pool.
    for batch in pool.iterate_batches():
        assert ListBytesBatcher.size(batch) <= batch_size

    # Convert batches back to list of items.
    items_flatten = [item for batch in pool.iterate_batches() for item in batch]

    # Compare initial list of items with flatten one.
    assert items == items_flatten, items_flatten



import json

from b_aws_testing_framework.credentials import Credentials
from botocore.response import StreamingBody

from b_lambda_layer_common_test.integration.infrastructure.main_stack import MainStack


def test_RESOURCE_lambda_layer_WITH_deployed_lambda_function_1_EXPECT_execution_successful():
    """
    Test whether the layer provides necessary functionality.

    :return: No return.
    """
    # Create client for lambda service.
    lambda_client = Credentials().boto_session.client('lambda')

    # Invoke specific lambda function.
    response = lambda_client.invoke(
        FunctionName=MainStack.get_output(MainStack.LAMBDA_FUNCTION_1_NAME_KEY),
        InvocationType='RequestResponse'
    )

    # Parse the result.
    payload: StreamingBody = response['Payload']
    data = [item.decode() for item in payload.iter_lines()]
    data = json.loads(''.join(data))

    # Assert that the result is as expected.
    assert data['Boto3Version'] == '1.16.35', data
    assert data['BotocoreVersion'] == '1.19.35', data


def test_RESOURCE_lambda_layer_WITH_deployed_lambda_function_2_EXPECT_execution_successful():
    """
    Test whether the layer provides necessary functionality.

    :return: No return.
    """
    # Create client for lambda service.
    lambda_client = Credentials().boto_session.client('lambda')

    # Invoke specific lambda function.
    response = lambda_client.invoke(
        FunctionName=MainStack.get_output(MainStack.LAMBDA_FUNCTION_2_NAME_KEY),
        InvocationType='RequestResponse'
    )

    # Parse the result.
    payload: StreamingBody = response['Payload']
    data = [item.decode() for item in payload.iter_lines()]

    print(data)


def test_RESOURCE_lambda_layer_WITH_deployed_lambda_function_3_EXPECT_execution_successful():
    """
    Test whether the layer provides necessary functionality.

    :return: No return.
    """
    # Create client for lambda service.
    lambda_client = Credentials().boto_session.client('lambda')

    # Invoke specific lambda function.
    response = lambda_client.invoke(
        FunctionName=MainStack.get_output(MainStack.LAMBDA_FUNCTION_3_NAME_KEY),
        InvocationType='RequestResponse'
    )

    # Parse the result.
    payload: StreamingBody = response['Payload']
    data = [item.decode() for item in payload.iter_lines()]
    data = json.loads(''.join(data))

    # Assert that the result is as expected.
    assert data['Boto3Version']
    assert data['BotocoreVersion']

def test_RESOURCE_lambda_layer_WITH_deployed_lambda_function_4_EXPECT_execution_successful():
    """
    Test whether a layer provides necessary functionality.

    :return: No return.
    """
    # Create client for Lambda service.
    lambda_client = Credentials().boto_session.client('lambda')

    # Invoke a specific lambda function.
    response = lambda_client.invoke(
        FunctionName=MainStack.get_output(MainStack.LAMBDA_FUNCTION_4_NAME_KEY),
        InvocationType='RequestResponse',
        Payload=json.dumps({'heartbeat': True})
    )

    # If Lambda function was not executed, it returns a string 'null'.
    data = response['Payload'].read().decode()
    assert data == 'null'



import re

from datetime import datetime
from typing import Optional

from b_lambda_layer_common.exceptions.container.bad_request_error import BadRequestError


class FieldValidator:
    """
    Validates various values.
    """
    @staticmethod
    def max_length(string_field: Optional[str], max_length: int, allow_null_value: bool = True) -> None:
        """
        Checks whether the string is not too long.

        :param string_field: Field to check.
        :param max_length: Max length.
        :param allow_null_value: Allow string field to be null and not check max length.

        :return: No return.
        """
        if allow_null_value:
            if string_field is None:
                return

        if len(string_field) > max_length:
            raise BadRequestError(
                f'Field with value "{string_field}" is too long. '
                f'Max length is {max_length}.'
            )

    @staticmethod
    def country_iso_3(string_field: Optional[str], allow_null_value: bool = True):
        """
        Checks whether the string complies with ISO3.

        :param string_field: Field to check.
        :param allow_null_value: Allow string field to be null and not check the ISO3 standard.

        :return: No return.
        """
        if allow_null_value:
            if string_field is None:
                return

        if len(string_field) != 3:
            raise BadRequestError(f'Field with value "{string_field}" does not comply with ISO3 standard.')

    @staticmethod
    def string_date(string_field: Optional[str], allow_null_value: bool = True):
        """
        Checks whether the string complies with standard date format.

        :param string_field: Field to check.
        :param allow_null_value: Allow string field to be null and not check the date format.

        :return: No return.
        """
        if allow_null_value:
            if string_field is None:
                return
        try:
            datetime.fromisoformat(string_field)
        except ValueError as ex:
            raise BadRequestError(str(ex))

    @staticmethod
    def string_email(string_field: Optional[str], allow_null_value: bool = True):
        if allow_null_value:
            if string_field is None:
                return

        regex = r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b'

        if not re.fullmatch(regex, string_field):
            raise BadRequestError(f'Field with value "{string_field}" is not an email.')



import json
import uuid
from typing import Any, Dict, Optional, Union
from json.decoder import JSONDecodeError

import boto3


class SfnCall:
    def __init__(self, state_machine_arn: str):
        self.__sfn_client = boto3.client('stepfunctions')
        self.__state_machine_arn = state_machine_arn

    def call(self, data: Dict[Any, Any]) -> Union[str, Dict[str, Any]]:
        response = self.__sfn_client.start_sync_execution(
            stateMachineArn=self.__state_machine_arn,
            name=str(uuid.uuid4()),
            input=json.dumps(data)
        )

        response_status = response.get('status')
        response_error = response.get('error')

        if response_status != 'SUCCEEDED':
            raise ValueError(
                f'Failed to call step functions state machine: {response_error}. '
                f'Read logs for more info.'
            )

        output = json.loads(response.get('output', {}))
        # Add http-like handling that does not break anything.
        output = self.__http_like_handling(output)

        return output

    @staticmethod
    def __http_like_handling(output: Dict[Any, Any]) -> Union[str, Dict[str, Any]]:
        """
        A function that treats responses/outputs from step functions state machine as an http response:
        for example, it looks for keys like "http_status" or "body".

        :param output: The output data from step functions state machine.

        :return: Original output, if the output does not contain http-like features
            Body of the output otherwise (JSON deserialized, if possible).
        """
        http_status: Optional[int] = output.get('http_status') or output.get('statusCode')

        if isinstance(body := output.get('body'), str):
            try:
                http_body = json.loads(body)
            except (JSONDecodeError, TypeError):
                http_body = body
        else:
            http_body = body

        if isinstance(http_status, int):
            if http_status >= 400:
                try:
                    error_message = json.dumps(http_body)
                except (JSONDecodeError, TypeError):
                    error_message = http_body

                raise ValueError(error_message)

            return http_body

        return output



from typing import Optional, Any

from pynamodb.attributes import UnicodeAttribute


class KmsAttribute(UnicodeAttribute):
    def __init__(self, kms_boto_client: Any, kms_arn: str, *args, **kwargs):
        self.__kms_arn = kms_arn
        self.__kms_boto_client = kms_boto_client

        super().__init__(*args, **kwargs)

    def serialize(self, value: Optional[str] = None) -> Optional[str]:
        if value:
            encrypted = self.__encrypt(value).decode()
            return super().serialize(encrypted)

    def deserialize(self, value: Optional[str] = None) -> Optional[str]:
        if value:
            decrypted = self.__decrypt(value.encode())
            return super().deserialize(decrypted)

    def __encrypt(self, sensitive_data: str) -> bytes:
        return self.__kms_boto_client.encrypt(
            KeyId=self.__kms_arn,
            Plaintext=sensitive_data.encode()
        )['CiphertextBlob']

    def __decrypt(self, sensitive_data: bytes) -> str:
        return self.__kms_boto_client.decrypt(
            KeyId=self.__kms_arn,
            CiphertextBlob=sensitive_data
        )['Plaintext'].decode()



from typing import Optional

from b_lambda_layer_common.exceptions.http_exception import HttpException


class AlreadyExistsError(HttpException):
    def __init__(self, message: Optional[str] = None):
        super().__init__(message)

    @staticmethod
    def http_code() -> int:
        return 409

    @staticmethod
    def identifier() -> str:
        return 'B_ALREADY_EXISTS'

    @staticmethod
    def description() -> str:
        return (
            'The server understood the request to create entities. However, the request can not be fulfilled because '
            'there are already existing entities and overriding them is not possible or not preferable. Usually, '
            'when a client gets this type of error, an "update" request should be sent, instead of "create".'
        )



from aws_cdk.core import Construct
from b_aws_testing_framework.tools.cdk_testing.testing_stack import TestingStack

from b_lambda_layer_common_test.integration.infrastructure.function1 import Function1
from b_lambda_layer_common_test.integration.infrastructure.function2 import Function2
from b_lambda_layer_common_test.integration.infrastructure.function3 import Function3
from b_lambda_layer_common_test.integration.infrastructure.function4 import Function4
from b_lambda_layer_common_test.integration.infrastructure.function_with_unit_tests import FunctionWithUnitTests


class MainStack(TestingStack):
    LAMBDA_FUNCTION_1_NAME_KEY = 'LambdaFunctionName1'
    LAMBDA_FUNCTION_2_NAME_KEY = 'LambdaFunctionName2'
    LAMBDA_FUNCTION_3_NAME_KEY = 'LambdaFunctionName3'
    LAMBDA_FUNCTION_4_NAME_KEY = 'LambdaFunctionName4'

    LAMBDA_FUNCTION_UNIT_TESTS_NAME_KEY = 'LambdaFunctionNameUnitTests'

    def __init__(self, scope: Construct):
        super().__init__(scope=scope)

        self.f1 = Function1(self)
        self.f2 = Function2(self)
        self.f3 = Function3(self)
        self.f4 = Function4(self)

        self.f_unit_tests = FunctionWithUnitTests(self)

        self.add_output(self.LAMBDA_FUNCTION_1_NAME_KEY, value=self.f1.function_name)
        self.add_output(self.LAMBDA_FUNCTION_2_NAME_KEY, value=self.f2.function_name)
        self.add_output(self.LAMBDA_FUNCTION_3_NAME_KEY, value=self.f3.function_name)
        self.add_output(self.LAMBDA_FUNCTION_4_NAME_KEY, value=self.f4.function_name)

        self.add_output(self.LAMBDA_FUNCTION_UNIT_TESTS_NAME_KEY, value=self.f_unit_tests.function_name)
